#include "fedcli_proc.h"

//----------------------------------------------------------------------
void InitCommandSet(Tcommandset &commandset, Tcommandtable &commandtable)
{
	commandset["help"]       = "[<command>] : show command hel or command list";
	commandtable["help"] = CLICOMM_SHOW_HELP;
	commandset["shell"]      = "escape to the shell (type exit to return to fed)";
	commandtable["shell"] = CLICOMM_SHELL;
	commandset["get"]        = "<remotefile> <localfile> : download a remote file";   
	commandtable["get"] = CLICOMM_GET;
	commandset["put"]        = "<localfile> <remotefile> : upload a local file";
	commandtable["put"] = CLICOMM_PUT;
	commandset["list"]       = "retrieve availeble file list";
	commandtable["list"] = CLICOMM_LIST;
	commandset["showlist"]   = "show current file list";
	commandtable["showlist"] = CLICOMM_SHOW_LIST;
	commandset["update"]     = "<remotefile> : execute a sort of remote update (actually dummy)";
	commandtable["update"] = CLICOMM_UPDATE;
	commandset["delete"]     = "<remotefile> : remove a file from fed system";
	commandtable["delete"] = CLICOMM_DELETE;
	commandset["disconnect"] = "close all active connections end leave fed";
	commandtable["disconnect"] = CLICOMM_DISCONNECT;
	commandset["lock"]       = "<remotefile> : lock a remote file";
	commandtable["lock"] = CLICOMM_LOCK;
}
//-------------------------------------------------------------------------
SOCKDESCR GetServerConnection(SOCKDESCR sockfd, SA_IN &addr, Tservtable servtable)
{
	string failedserver_ipaddr;
	int    failedserver_port;
	
	if(sockfd != -1){
		failedserver_ipaddr = GetIPAddr(sockfd,PEER_ENDPOINT);
		failedserver_port   = GetIPPort(sockfd,PEER_ENDPOINT);
		Close(sockfd);
	} else {
		failedserver_ipaddr = "tasso";
		failedserver_port = -1;
	} 
	
	bool done = false;
	for(unsigned int j = 0; (j < servtable.size()) && (!done); j++){
		if( (sockfd = Socket(AF_INET,SOCK_STREAM,0)) == -1){
			perror("Client CANNOT start up");
			return -1;
		}
		string currservaddr = servtable[j].ip_addr;
		int    currservport = servtable[j].port;
#ifdef _DEBUG
cout<<"GetServerConnectio: "<<currservaddr<<":"<<currservport<<endl;
#endif
		//ovviamente non proviamo a connetterci a un server appena failed
		if((currservaddr != failedserver_ipaddr) && (currservport != failedserver_port)){
			inet_pton(AF_INET,currservaddr.c_str(),&addr.sin_addr);
			addr.sin_port = htons(currservport);
#ifdef _DEBUG
cout<<"trying to connect to port: "<<servtable[j].port<<endl;
#endif
			if( (sockfd = Reconnect(sockfd,(SA *) &addr,sizeof(addr),MAX_RESUME_TIME)) != -1) done = true;
			else Close(sockfd);
		}
	}
	if(!done){
		cout<<"I CANNOT CONNECT TO ANY SERVER. Exiting..."<<endl;
		return -1;
	}else{
		cout<<"CONNECTED to server: "<<GetIPAddr(sockfd,PEER_ENDPOINT)<<":"<<GetIPPort(sockfd,PEER_ENDPOINT)<<endl;
		return sockfd;
	}
}
//---------------------------------------------------------------
void ShowCommandSet(Tcommandset commandset)
{
	map<string,string>::iterator iter = commandset.begin();
	while(iter != commandset.end()){	
		int spacetoprint = COMMAND_SPACE_TO_PRINT - iter->first.size();
		string space(spacetoprint,' ');
		cout<<iter->first<<space<<iter->second<<endl;
		iter++;
	}
	cout<<endl;
}
//-------------------------------------------------------------------------
void ShowCommandSet(Tcommandset commandset, string command)
{
	cout<<command<<" "<<commandset[command]<<endl;
}
//-------------------------------------------------------------------------
void ShowFileList(const Tstringlist &filelist)
{
	if(filelist.empty()){
		cerr<<"file list is empty!"<<endl;
		return;
	}
	int res = 0;
	for(unsigned int j = 0; j < filelist.size(); j++){
		int spacetoprint = FILE_SPACE_TO_PRINT - filelist[j].size();
		string space(spacetoprint,' ');
		cout<<filelist[j]<<space;
		if(res == 2){
			res = 0;
			cout<<endl;
		}
		else res++;
	}
}
//--------------------------------------------------------------------------			
//refresh list
// 0 success
//-1 something was wrong
int RefreshFileList(Tstringlist &filelist)
{
	//verifichiamo che il file esista veramente
	ifstream input(CLIENT_FILELIST_TEMP_NAME);
	string name;
	input>>name;
	if(name != FILEDB_BEGIN_TAG){
		cerr<<"bad BOF in temp filename"<<endl;
		exit(-1);
		return -1;
	}
	//vector temporaneo
	Tstringlist templist;
	int checkguard = 0;
	input>>name;
	while(name != FILEDB_END_TAG){
		templist.push_back(name);
		input>>name;
		if(checkguard++ > MAX_GUARD){
			cerr<<"filelist exeeded maximum length."<<endl;
			exit(-1);
			return -1;
		}
	}
	//ok lista buona;
	filelist = templist;
	return 0;
}		    
//--------------------------------------------------------------------
//--------------------------------------------------------------------
//esegue il comando list
//restituisce:
// 0 in caso di successo
//-1 in caso di DENIED
//-2 in caso di redirect
//-3 download truncato ripetere...
//-4 errori irrecuperabili

int ExecuteListCommand(SOCKDESCR sockfd)
{
	//il server e' in attesa di un comando
	string msg_to_send;
	PrepareMessage(CLICOMM_LIST,msg_to_send);
	if(SendMsg(sockfd,msg_to_send) == -1){
		cerr<<"Unable to deliver command. Aborting/redirecting connection"<<endl;
		return -2;
	}
	string msg_received;
 	if(RecvMsg(sockfd,msg_received) == -1){
		cerr<<"Unable to receive server answer: Aborting"<<endl;
		return -2;
	}
#ifdef _DEBUG
cout<<"list command recv msg: "<<msg_received<<endl;
#endif
	//messaggio ricevuto correttamente: processiamolo
	Trepl repl;
	ProcessMessageForClient(CLICOMM_LIST,msg_received,repl);
	switch (repl.msg_type){
		case ACK:{
			string servaddr = GetIPAddr(sockfd,PEER_ENDPOINT);
			int FTresult = FTClientProcedure(servaddr,repl.port,CLIENT_FILELIST_TEMP_NAME,repl.filelen,OVERWRITE);
			switch(FTresult){
				case 0: //tutto e' bene quel che finisce bene
					return 0;
				break;
				
				case -1:{ //impossibile aprire il filetemporaneo di list
					cerr<<"impossibile creare il file temporaneo per il filelist"<<endl;
					return -4;
				}
				break;
				
				case -2:{ //il client non riesce a fare la connect
					cout<<"Server refuse connection for download"<<endl;
					return -2;
				} 
				break;
				
				case -3: //download troncato
					return -3;
				break; 
				
				default:{
					cerr<<"codice di errore non contemplato"<<endl;
					return -4;
				}
			}
		}
		break;
		
		case DENY:
			//il severver ha risposto con un divieto di accesso.
			return -1;
		break;
		
		default:
			return -2;
	}
}
//-----------------------------------------------------------------						
//esegue il comando di download
//  0 tutto OK
// -1 remotefile *non* e' nella lista;
// -2 download andato male perche' il server non risponde
// -3 download troncato
// -4 denied
// -5 impossibile aprire localmente il file
// -6 REDIRECT
// -7 altri
int ExecuteGetCommand(SOCKDESCR sockfd, const Tstringlist &filelist, string remotefile, string localfile, string &redir_ipaddr, int &redir_ipport)
{
	string servaddr = GetIPAddr(sockfd,PEER_ENDPOINT);
	//int serverport  = GetIPPort(sockfd,PEER_ENDPOINT);
	
	//prima cosa: vedere se remote file e' esportato
	if(find(filelist.begin(),filelist.end(),remotefile) == filelist.end())
		return -1;
	
	//ok il file e' nella lista ora provo a fare un download al server indicato
	string msg_to_send;
	PrepareMessage(CLICOMM_GET,remotefile,msg_to_send); 

#ifdef _DEBUG
cout<<"ExecuteGetCommand: sending message"<<endl;
#endif
		
	if(SendMsg(sockfd,msg_to_send) == -1){
		cerr<<"Unable to deliver command. Aborting"<<endl;
		//la SendMsg *riprova* a trasmettere piu' volte
		//se fallisce si dichiara la connessione decaduta
		return -2;
	}
	string msg_received;
 
#ifdef _DEBUG
cout<<"ExecuteGetCommand: receiving message"<<endl;
#endif

 	if(RecvMsg(sockfd,msg_received) == -1){
		cerr<<"Unable to receive server answer: Aborting"<<endl;
		return -2;
	}
	//messaggio ricevuto correttamente: processiamolo
	Trepl repl;
	ProcessMessageForClient(CLICOMM_GET,msg_received,repl);
	switch (repl.msg_type){
		case ACK:{
			int FTresult = FTClientProcedure(servaddr,repl.port,localfile,repl.filelen,OVERWRITE);
			switch(FTresult){
				case 0: //ok, download con successo
					return 0;
				break;
				
 				case -1: //impossibile aprire un file locale
					return -5;
				break;
				
				case -2: //connect failed redirect?
					return -2;
				break;
				
				case -3: //download incompleto
					return -3;
				break;
						
				default:
					return -7;
			}
		}
		break;
		
		case DENY:
			return -4;
		break;
		
		case REDIR:{ //il server ha risposto con REDIRECT
			//settiamo le variabili statiche di redirect
			redir_ipaddr = repl.ipaddr;
			redir_ipport = repl.port;
			//bool done = false;
			//while(!done){
			SA_IN redserver;
			memset(&redserver,0,sizeof(redserver));
			redserver.sin_family = AF_INET;
			redserver.sin_port   = htons(redir_ipport);
			inet_pton(AF_INET,redir_ipaddr.c_str(),&redserver.sin_addr);
			
			SOCKDESCR redirsockfd = Socket(AF_INET,SOCK_STREAM,0);
			if( (redirsockfd = Reconnect(redirsockfd,(SA *)&redserver,sizeof(redserver),MAX_RESUME_TIME)) == -1){
				cout<<"unable to connect to redirected server. Aborting DL"<<endl;
				return -2;
			}
			//ok collegato!
			//adesso chiamo la procedure ExecuteDL
			int redires = ExecuteGetCommand(redirsockfd,filelist,remotefile,localfile,redir_ipaddr,redir_ipport);		
			string byebyemsg;
			PrepareMessage(CLICOMM_DISCONNECT,byebyemsg);
			SendMsg(redirsockfd,byebyemsg);
			return redires;
		}break;
		default:{
			cerr<<"message unrecognized"<<endl;
			return -5;
		}
	}	
}
//---------------------------------------------------------------------
//0 in caso di successo
//-1 in caso di denay dell' owner
//-2 uncatched errors
int updateredir(string filename, string redir_addr, int redir_port)
{
	// attivo il ciclo con cui rincorro l'owner
	SA_IN redirserver;
	while(true){
		memset(&redirserver,0,sizeof(redirserver));
		redirserver.sin_family = AF_INET;
		redirserver.sin_port = htons(redir_port);
		inet_pton(AF_INET,redir_addr.c_str(),&redirserver.sin_addr);
				
		SOCKDESCR redirsock = Socket(AF_INET,SOCK_STREAM,0);;
		if( (redirsock = Reconnect(redirsock,(SA *)&redirserver,sizeof(redirserver),MAXRESUME)) == -1){
			cerr<<"updateredir: cannoct connect to next server: bouncing chain interrupted"<<endl;
			return -2;
		}
		//ok sono collegato
		string msgtosend;
		PrepareMessage(CLICOMM_UPDATE,filename,msgtosend);
		if(SendMsg(redirsock,msgtosend) == -1){
			cerr<<"updateredir: unable to deliver redir msg. Aborting"<<endl;
			return -2;
		}
		string recvmsg;
		if(RecvMsg(redirsock,recvmsg) == -1){
			cerr<<"updateredir: unable to receive redir reply. Aborting"<<endl;
			return -2;
		}
		Trepl repl;
		ProcessMessageForClient(CLICOMM_UPDATE,recvmsg,repl);	
		switch(repl.msg_type){
			case ACK:{
				//chiudo la connessione:
				string discmsg;
				PrepareMessage(CLICOMM_DISCONNECT,discmsg);
				SendMsg(redirsock,discmsg);
				Close(redirsock);
				return 0;
			}break;
			case DENY:
				return -1;
			break;
			case REDIR:{	
				string discmsg;
				PrepareMessage(CLICOMM_DISCONNECT,discmsg);
				SendMsg(redirsock,discmsg);
				Close(redirsock);
				redir_addr = repl.ipaddr;
				redir_port = repl.port;
			}break;
			default:{
				cerr<<"updatredr: messagio non riconosciuto. aborting"<<endl;
				return -2;
			}
		}
	}//fine forever
}
//----------------
// 0 update ok
//-1 file locked no update allowed
//-2 unrecoverable errors no update allowed
//-3 file non nella lista
int ExecuteUpdateCommand(SOCKDESCR connsock,const Tstringlist &filelist, string filename)
{
	//solo al primo si da il consenso al fetch on write
	//tutti i rimanenti bounced server non sono chiamati al
	//caching
	if(find(filelist.begin(),filelist.end(),filename) == filelist.end())
		return -3;
	string msgtosend;
	PrepareMessage(CLICOMM_UPDATE_WITH_CACHE,filename,msgtosend);
	if(SendMsg(connsock,msgtosend) == -1){
		cerr<<"ExecuteUpdateCommand unable to deliver first msg. Aborting"<<endl;
		return -2;
	}
	string recvmsg;
	if(RecvMsg(connsock,recvmsg) == -1){
		cerr<<"ExecuteUpdateCommand: unable to receive first reply. Aborting"<<endl;
		return -2;
	}
	Trepl repl;
	ProcessMessageForClient(CLICOMM_UPDATE,recvmsg,repl);	
	switch(repl.msg_type){
		case ACK:
			//il server e' owner e ha aggiornato
			return 0;
		break;
		case DENY:
			//il server e' owner ma il file e' locked
			return -1;
		break;
		case REDIR:{
			//il server NON e' owner
			int redirresult = updateredir(filename,repl.ipaddr, repl.port);
			switch(redirresult){
				case 0:{ //ok, l'owner e' stato informato;
					//adesso prepariamo il messaggio di ACK al primo server ed usciamo
					string ackmsg;
					PrepareMessage(SERV_SIMPLE_ACK,ackmsg);
					SendMsg(connsock,ackmsg);
					return 0;
				}break;
				case -1:{  //owner risponde con deny
					string denymsg;
					PrepareMessage(SERV_DENY,denymsg);
					SendMsg(connsock,denymsg);
					return -1;
				}break;
				default:{
					cerr<<"ExecuteUpdateCommand:unrecoverable error:Aborting"<<endl;
					return -2;
				}
			}
		}break;
		default:{
			cerr<<"ExecuteUpdatecommand:errore: messaggio di risposta non contemplato"<<endl;
			return -2;
		}
	}
}
//--------------------------------------------------------------------------
	