
#include "fedserv.h"

Tfiledb   filedb(FILEDB_DEFAULT_NAME);
Tserverdb serverdb(SERVERDB_DEFAULT_NAME);

//mutexes:
pthread_mutex_t filedb_mutex   = PTHREAD_MUTEX_INITIALIZER;
pthread_mutex_t serverdb_mutex = PTHREAD_MUTEX_INITIALIZER;
 
bool csshutdown = false;
bool ssshutdown = false;

//--------------------------------------------------------
bool InitServer(void)
{
	if(serverdb.Size() == 0){
		cerr<<"UNABLE TO FIND server data base file ("<<SERVERDB_DEFAULT_NAME<<"). Server CANNOT start"<<endl;
		return false;
	}
	if(filedb.Size() == 0){
		cout<<"File data base ("<<FILEDB_DEFAULT_NAME<<") corrupted. Need a server lookup"<<endl;
		//qui andreabbe  gestita una fase di connect e douwnload del
		//filedb, esattamente come se fosse un client.
		//lo implementiamo se rimane tempo
		return false;
	}
	return true;
}	
//------------------------------------------------------------
bool CloseServer(void)
{
	cerr<<"NOT YET IMPLEMENTED"<<endl;
	return false; 
}
//----------------------------------------------------------
void *CSMainThread (void *arg)
{
	//adesso ricavo da arg le informazioni salienti
	SOCKDESCR listensock  = *((SOCKDESCR *)arg);
	delete arg;
	
	SA_IN cliaddr;
	 
	//OK, adesso genero il cicloinfinito
	while(!csshutdown){
		socklen_t cliaddrlen = sizeof(cliaddr);
		SOCKDESCR connsock = Accept(listensock, (SA *)&cliaddr, &cliaddrlen);
		//adesso creo un thread per ciascun client connesso
		SOCKDESCR *pconnsock = new(SOCKDESCR);
		*pconnsock = connsock;
		pthread_t csptid;
		pthread_create(&csptid,NULL,ClientToServer,(void *)pconnsock);
		//non faccio il join perche' i ClientToserver threads sono detached
	}
	//qui e' giunto l' ordine di abbandonare la nave:
	Close(listensock);
	pthread_exit(NULL);    
}
//--------------------------------------------------------------
void *SSMainThread(void *arg)
{
	//adesso ricavo da arg le informazioni salienti
	SOCKDESCR listensock = *((SOCKDESCR *)arg);
	delete arg;
	
	SA_IN remoteaddr;
	
	//ok adesso metto su il ciclo
	while(!ssshutdown){
		socklen_t remoteaddrlen;
		SOCKDESCR connsock = Accept(listensock,(SA *)&remoteaddr,&remoteaddrlen);
		//ok qui oh un server connesso
		//devo gestire i vari protocolli. Il server qui e' seriale
		string commandmsg;
		if(RecvMsg(connsock,commandmsg) == -1){
			cerr<<"unable to receive msg from remote server. shutdown connection"<<endl;
			Close(connsock);
		}else{
			Tclicomm servercomm;
			ProcessMessageForServer(SERVER_COMMAND,commandmsg,servercomm);
			switch(servercomm.msg_type){
				case PSO_MSG:{
					istringstream is(servercomm.arg2);
					int num; is>>num;
#ifdef _DEBUG
cout<<"Server: ManagePSO..."<<endl;
#endif
					ManagePSO(connsock,servercomm.arg1,num);
				}
				break;
				case CAC_INV:{
#ifdef _DEBUG
cout<<"Server-to-server: managing incoming CACHE_INV"<<endl;
#endif
					string filename = servercomm.arg1;
					pthread_mutex_lock(&filedb_mutex);
					if(filedb.GetFileLocation(filename) == CACHED){
						//sono sicuro che la cache_invalidate arriva dal solo owner
						string newowner = GetIPAddr(connsock,PEER_ENDPOINT);
						filedb.SetFileOwner(filename,newowner);
					}
					pthread_mutex_unlock(&filedb_mutex);
					string simpleackmsg;
					PrepareMessage(SERV_SIMPLE_ACK,simpleackmsg);
					SendMsg(connsock,simpleackmsg);
					Close(connsock);
				}break;
				default:{
					cerr<<"SERVER TO SERVER COMMAND NOT RECOGNIZED"<<endl;
					Close(connsock);
				}
			}
		}
	}
	//ok abbandoniamo la nave:
	Close(listensock);
	pthread_exit(NULL);
} 
//----------------------------------------------------------------------
int main(int argc, char **argv)
{
	cout<<"starting server"<<endl;
	
	if(!InitServer()){
		cerr<<"Il server non puo' partire. Aborting..."<<endl;
		exit(-1);
	}
	cout<<"INIT PHASE successfully completed"<<endl;
	
	SA_IN csaddr,ssaddr;
	
	
	memset(&csaddr,0,sizeof(csaddr));
	csaddr.sin_family     = AF_INET;
	csaddr.sin_port       = SERVER_FOR_CLIENT_PORT;  //porta di ascolto Client->Server
	csaddr.sin_addr.s_addr = htonl(INADDR_ANY);
	
	memset(&ssaddr,0,sizeof(ssaddr));
	ssaddr.sin_family     = AF_INET;
	ssaddr.sin_port       = SERVER_FOR_SERVER_PORT;  //porta di ascolto Server -> Server
	ssaddr.sin_addr.s_addr = htonl(INADDR_ANY);
	
	if(argc > 2){
		cout<<"reimposting client listen port (default 45000) to "<<argv[1]<<endl;
		csaddr.sin_port = htons(atoi(argv[1]));
		cout<<"setting server listen port (default 50000) to "<<argv[2]<<endl;
		ssaddr.sin_port = htons(atoi(argv[2]));
	}
	
	SOCKDESCR cslistensock = Socket(AF_INET,SOCK_STREAM,0);
	SOCKDESCR sslistensock = Socket(AF_INET,SOCK_STREAM,0);
	
	if(!Bind(cslistensock,(SA *)&csaddr,sizeof(csaddr))){
		cerr<<"Unable to continue. Exiting"<<endl;

		exit(-1);
	}
	if(!Bind(sslistensock,(SA *)&ssaddr,sizeof(ssaddr))){
		cerr<<"Ynable to continue. Exiting"<<endl;
		exit(-1);
	}
	if(!Listen(cslistensock,MAXLISTEN)){
	cerr<<"Unable to continue. Exiting"<<endl;
		exit(-1);
	}
	if(!Listen(sslistensock,MAXLISTEN)){
	cerr<<"Unable to continue. Exiting"<<endl;
		exit(-1);
	}
	
	//adesso genero i due main thread
	pthread_t cstid,sstid;
	
	SOCKDESCR *psslistensock = new(SOCKDESCR);
	*psslistensock = sslistensock;
	pthread_create(&sstid,NULL,SSMainThread,(void *)psslistensock);
	
	SOCKDESCR *pcslistensock = new(SOCKDESCR);
	*pcslistensock = cslistensock;
	pthread_create(&cstid,NULL,CSMainThread,(void *)pcslistensock);
	
	//fine main thread metto dei join per estetica
	pthread_join(sstid,NULL);
	pthread_join(cstid,NULL);
	
	if(!CloseServer())
		cout<<"Il server non ha eseguito correttamente la fase di shutdown"<<endl;
	else
		cout<<"Server is going DOWN. See ya soon"<<endl;
} 
//----------------------------------------------------------------------------------