#include "processfile.h"


//implementazione delle file processing facilities
/**********
* apre il filedb e carica tutte le linee in esso contenute
* RETURN VALUE:
* TRUE se il caricamento e' andato abuon fine
* FALSE se problemi vari (open fallisce etc)
***********/
bool Tfiledb::_loadfrom(void)
{
#ifdef _DEBUG
cout<<"starting readin file.db: "<<filedbname<<endl;
#endif
	ifstream inputfile(filedbname.c_str());
	string begintag = "none";
	string endtag = "none";
	inputfile>>begintag;
	if(begintag != FILEDB_BEGIN_TAG){
		cerr<<"filedb does NOT exist or corrupted"<<endl;	
		return false;
	}
	filerecord record;
	inputfile>>record.filename;
	while(record.filename != FILEDB_END_TAG){
		inputfile>>record.status;
		inputfile>>record.location;
		inputfile>>record.numofaccess;
		inputfile>>record.owner;
		recordlist.push_back(record);
		inputfile>>record.filename;
	}
	return true;	
}
//----------------------------------------------
/********************
* compila il file.db
********************/
bool Tfiledb::_writeto(void)
{
	//apro l'ofstream
	//SOVRASCRIVO SEMPRE!
	ofstream outfile(filedbname.c_str());
	//scrivo il beginnin tag
	outfile<<FILEDB_BEGIN_TAG<<"\n";
	//cicliamo
	for(unsigned int j = 0; j < recordlist.size(); j++){
		int spaceleft = FIRST_SPACE - recordlist[j].filename.size();
		string space(spaceleft,' ');		
		outfile<<recordlist[j].filename<<space<<recordlist[j].status<<SPACE<<recordlist[j].location<<SPACE<<recordlist[j].numofaccess<<SPACE<<recordlist[j].owner<<"\n";
	}
	outfile<<FILEDB_END_TAG<<"\n";
	return true;
}
bool Tfiledb::_writeto(string copyfilename)
{
	//apro l'ofstream
	//SOVRASCRIVO SEMPRE!
	ofstream outfile(copyfilename.c_str());
	//scrivo il beginnin tag
	outfile<<FILEDB_BEGIN_TAG<<"\n";
	//cicliamo
	for(unsigned int j = 0; j < recordlist.size(); j++){
		int spaceleft = FIRST_SPACE - recordlist[j].filename.size();
		string space(spaceleft,' ');		
		outfile<<recordlist[j].filename<<space<<recordlist[j].status<<SPACE<<recordlist[j].location<<SPACE<<recordlist[j].numofaccess<<SPACE<<recordlist[j].owner<<"\n";
	}
	outfile<<FILEDB_END_TAG<<"\n";
	return true;
}
//----------------------------------------------
vector<filerecord>::iterator Tfiledb::_findname(string name)
{
	vector<filerecord>::iterator iter = recordlist.begin();
	while(iter != recordlist.end())
		if(iter->filename == name) return iter;
		else iter++;
	return iter;
}	
//-----------------------------------------------------------	
Tfiledb::Tfiledb(void)
{
	// crea l'oggetto senza caricare il db
	modified = false;
	filedbname = "";
}
//----------------------------------------------
Tfiledb::Tfiledb(string filename)
{
	modified = false;
	filedbname = filename;
	if(!_loadfrom()){
		cerr<<"impossibile caricare correttamente il filedb... aborting"<<endl;
		exit(-1);
	}
}
//---------------------------------------------------- 
bool Tfiledb::UpdateToDisk(void)
{
	//aggiorna il filedb sul disco
	//INPOSTA il flag MODIFIED A FALSE!
	
	if(!modified) return true;
	if(!_writeto()){
		cerr<<"ERRORE: impossibile salvare il filedb su disco. Continuo"<<endl;
		return false;
	}	
	else{
		modified = false;
		return true;
	}
}

bool Tfiledb::MakeCopyToDisk(string filename)
{
	if(!_writeto(filename)){
		cerr<<"ERRORE: impossibile copiare il filedb su disco. Continuo"<<endl;
		return false;
	}	
	else return true;
}
//------------------------------------------------------
int Tfiledb::GetFileStatus(string filename)
{
	vector<filerecord>::iterator iter = _findname(filename);
	if(iter == recordlist.end())
		return NONE;
	else
		return iter->status;
}
//--------------------------------------------------------
int Tfiledb::GetFileLocation(string filename)
{
	vector<filerecord>::iterator iter = _findname(filename);
	if(iter == recordlist.end())
		return NONE;
	else
		return iter->location;
}
//--------------------------------------------------------
string Tfiledb::GetFileOwner(string filename)
{
	vector<filerecord>::iterator iter = _findname(filename);
	if(iter == recordlist.end())
		return "";
	else
		return iter->owner;
}

//--------------------------------------------------------
int Tfiledb::GetNumOfAccess(string filename)
{
	vector<filerecord>::iterator iter = _findname(filename);
	if(iter == recordlist.end())
		return -1;
	else
		return iter->numofaccess;
}
//--------------------------------------------------------
bool Tfiledb::IsFileLocked(string filename)
{
	vector<filerecord>::iterator iter = _findname(filename);
	if(iter == recordlist.end())
		return false;
	else if(iter->status == LOCKED)
			return true;
		else return false;
}
//-----------------------------------------------------------
bool Tfiledb::SetFileLock(string filename)
{
	vector<filerecord>::iterator iter = _findname(filename);
	if(iter == recordlist.end())
		return false;
	else iter->status = LOCKED;
}
//-----------------------------------------------------------
bool Tfiledb::RemoveFileLock(string filename)
{
	vector<filerecord>::iterator iter = _findname(filename);
	if(iter == recordlist.end())
		return false;
	else iter->status = NONE;
}
//--------------------------------------------------------
bool Tfiledb::SetFileStatus(string filename, int newstatus)
{
	vector<filerecord>::iterator iter = _findname(filename);
	if(iter == recordlist.end()){
		cerr<<"file not found in recordlist"<<endl;
		return false;
	}
	switch(newstatus){
		case OPEN:
		case CLOSED:
		case LOCKED:
		case NONE:{
			iter->status = newstatus;
			modified = true;
			return true;
		}break;
		default:{
			cerr<<"filedb:new status unrecognized"<<endl;
			return false;
		}
	}
}
//----------------------------------------------------------------------
bool Tfiledb::SetFileLocation(string filename, int newlocation)
{
	vector<filerecord>::iterator iter = _findname(filename);
	if(iter == recordlist.end()){
		cerr<<"file not found in recordlist"<<endl;
		return false;
	}
	switch(newlocation){
		case OWNED:
		case CACHED:
		case REMOTE:{
			iter->location = newlocation;
			modified = true;
			return true;
		}break;
		default:{
			cerr<<"filedb: new location unrecognized"<<endl;
		 	return false;
		}
	}
}
//-----------------------------------------------------------
bool Tfiledb::SetFileOwner(string filename, string newaddr)
{
	vector<filerecord>::iterator iter = _findname(filename);
	if(iter == recordlist.end()){
		cerr<<"file not found in recordlist"<<endl;
		return false;
	}
	//nessun controllo ora?
	iter->owner = newaddr;
	modified = true;
	return true;
}
//------------------------------------------------------------------
int Tfiledb::IncrementNumOfAccess(string filename)
{
	vector<filerecord>::iterator iter = _findname(filename);
	if(iter == recordlist.end()){
		cerr<<"file not found in recordlist"<<endl;
		return false;
	}
	iter->numofaccess += 1;
	modified = true;
	return iter->numofaccess;
}
//------------------------------------------------------------------
bool Tfiledb::GetList(vector<string> &filelist) const
{
	filelist.clear();
	for(unsigned int j = 0; j < recordlist.size(); j++)
		filelist.push_back(recordlist[j].filename);
	sort(filelist.begin(),filelist.end());
	return true;
}
//-------------------------------------------------------------
bool Tfiledb::ExtractFileListToDisk(string filename) const
{
	ofstream of(filename.c_str());
	of<<FILEDB_BEGIN_TAG<<"\n";
	for(unsigned int j = 0; j < recordlist.size(); j++)
		of<<recordlist[j].filename<<"\n";
	of<<FILEDB_END_TAG<<"\n";
	return true;
}	
//-------------------------------------------------------------------
bool Tfiledb::Insert(string filename, int status, int location, string owner, int numofaccess)
{
	if(_findname(filename) != recordlist.end()){
		cout<<filename<<" is already in the database. No insertion allowed"<<endl;
		return false;
	} 	
	filerecord rec;
	rec.filename = filename; rec.status = status;
	rec.location = location; rec.owner = owner;rec.numofaccess = numofaccess;
	recordlist.push_back(rec);
	modified = true;
	return true;
}
//-------------------------------------------------------
bool Tfiledb::Find(string filename)
{
	if(_findname(filename) == recordlist.end())
		return false;
	else return true;
}
//------------------------------------------------------------
bool Tfiledb::Delete(string filename)
{
	vector<filerecord>::iterator iter = _findname(filename);
	if(iter == recordlist.end()){
		cerr<<filename<<"is NOT in the file database. Delete NOT allowed"<<endl;
		return false;
	}
	recordlist.erase(iter);
	modified = true;
	return true;
}  
//----------------------------------------------------------
int Tfiledb::Size() const
{
	return recordlist.size();
}
//-------------------------------------------------------
//------------------------------------------------------------
//------------------------------------------------------------
//------------------------------------------------------------
bool Tserverdb::_loadfrom(void)
{
	ifstream infile(serverdbname.c_str());
	string begintag = "none";
	string endtag   = "none";
	infile>>begintag;
	if(begintag != FILEDB_BEGIN_TAG){
		cerr<<"serverdb does not exists or corrupted"<<endl;
		return false;
	}
	serverrecord rec;
	infile>>rec.name;
	while(rec.name !=	FILEDB_END_TAG){
		infile>>rec.ipaddr;
		infile>>rec.port;
		infile>>rec.serverport;
		infile>>rec.status;
#ifdef _DEBUG
		cout<<"server.db: name,addr,port,serport,status ->"<<rec.name<<","<<rec.ipaddr<<","<<rec.port<<","<<rec.serverport<<","<<rec.status<<endl;
#endif
		recordvector.push_back(rec);
		infile>>rec.name;
	}
	return true;
}
//------------------
bool Tserverdb::_writeto(void)
{
	ofstream outfile(serverdbname.c_str());
	outfile<<FILEDB_BEGIN_TAG<<"\n";
	for(unsigned int j = 0; j < recordvector.size(); j++){
		string firstspace(FIRST_SPACE-recordvector[j].name.size(),' ');
		ostringstream os;
		os<<recordvector[j].port;
		string strport = os.str();
		int spaceleft = 5 - strport.size();
		string extraspace(spaceleft,' ');
		outfile<<recordvector[j].name<<firstspace<<recordvector[j].ipaddr<<SPACE<<extraspace<<strport<<SPACE<<recordvector[j].serverport<<SPACE<<recordvector[j].status<<"\n";
	}
	outfile<<FILEDB_END_TAG<<"\n";
	return true;
}
//-----------------------------------------------------------
serverrecord Tserverdb::Get(void)
{
	if((unsigned int)currentindex != recordvector.size())
	{
		return recordvector[currentindex++];
	}else{
		serverrecord temp;
		temp.ipaddr = ""; temp.port = 0; temp.status = 0;
		return temp;
	}
}
//---------------------------------------------------------------
int Tserverdb::SetPosition(int position)
{
	if((unsigned int)position < recordvector.size()){
		currentindex = position;
		return position;
	}else{
		currentindex = recordvector.size();
		return -1;
	}
}
//---------------------------------------------------------------------
Tserverdb::Tserverdb(void)
{
	modified = false;
	serverdbname = "";
}
//----------------------------------------------------------------
Tserverdb::Tserverdb(string filename)
{
	modified = false;
	serverdbname = filename;
	if(!_loadfrom()){
		cerr<<"impossobile caricare il database dei server. Aborting"<<endl;
		exit(-1);
	}
}
//--------------------------------------------------------------------------
int Tserverdb::Size(void)
{
	return recordvector.size();
}
//---------------------------------------------------------------------------
int Tserverdb::GetPort(string name) const
{
	for(unsigned int j = 0; j < recordvector.size(); j++)
		if(name == recordvector[j].name)
			return recordvector[j].port;
	return 0;
}
//-------------------------------------------------------------------------------
int Tserverdb::GetServerPort(string name) const
{
	for(unsigned int j = 0; j < recordvector.size(); j++)
		if(name == recordvector[j].name)
			return recordvector[j].serverport;
	return 0;
}	
//----------------------------------------------------------------------------
string Tserverdb::GetAddr(string name) const
{
	for(unsigned int j = 0; j < recordvector.size(); j++)
		if(name == recordvector[j].name)
			return recordvector[j].ipaddr;
	//return "wrong_serveraddr";
	return 0;
}
//----------------------------------------------------------------------------
string Tserverdb::GetNameFromAddr(string addr) const
{
	for(unsigned int j = 0; j < recordvector.size(); j++)
		if(addr == recordvector[j].ipaddr)
			return recordvector[j].name;
	//return "wrong_servername";
	return 0;
}
//----------------------------------------------------------------------------
int Tserverdb::Size() const
{
	return recordvector.size();
}
//--------------------------------------------------------------------------
//--------------------------------------------------------------------------
//--------------------------------------------------------------------------
bool FillServTable(Tservtable &servtable,string filename)
{
	if(!servtable.empty())	servtable.clear();
#ifdef _DEBUG
	cout<<"opening inputfile"<<endl;
#endif	
	ifstream inputfile(filename.c_str());
	string begintag = "none",endtag = "none";
	inputfile>>begintag;
	if(begintag != FILEDB_BEGIN_TAG){
		cerr<<"filedb does NOT exist or corrupted"<<endl;	
		return false;
	}
	servrecord record;
	inputfile>>record.ip_addr;
	while(record.ip_addr != FILEDB_END_TAG){
		inputfile>>record.port;
#ifdef _DEBUG
		cout<<"IP Addr: "<<record.ip_addr<<" port: "<<record.port<<endl;
#endif
		servtable.push_back(record);
		inputfile>>record.ip_addr;
	}
	return true;
}